# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from v3iofs import V3ioFile, V3ioFS


def test_fetch_range(fs: V3ioFS, tmp_obj):
    v3f = V3ioFile(fs, tmp_obj.path)
    start, end = 3, len(tmp_obj.data) - 3
    data = v3f._fetch_range(start, end)
    expected = tmp_obj.data[start:end]
    assert expected == data, "bad data"


def test_upload_chunk(fs: V3ioFS, tmp_obj):
    v3f = V3ioFile(fs, tmp_obj.path, "ab")
    chunk = b"::chunk of data"
    v3f.buffer.write(chunk)
    v3f._upload_chunk()

    expected = tmp_obj.data + chunk

    with fs.open(tmp_obj.path, "rb") as fp:
        data = fp.read()

    assert expected == data, "bad data"


def test_initiate_upload(fs: V3ioFS, tmp_obj):
    fs.touch(tmp_obj.path)
    assert fs.exists(tmp_obj.path)
    v3f = V3ioFile(fs, tmp_obj.path, "wb")
    v3f._initiate_upload()
    assert not fs.exists(tmp_obj.path)
    # should not fail even if the file does not exist
    v3f._initiate_upload()



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
_ignore_statuses = {200, 204, 206, 409}


def handle_v3io_errors(response, file_path):
    if response.status_code in _ignore_statuses:
        return response.body
    if response.status_code == 404:
        raise FileNotFoundError(f"{file_path!r} not found")
    raise Exception(f"{response.status_code} received while accessing {file_path!r}")



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid
from collections import namedtuple
from datetime import datetime
from getpass import getuser
from http import HTTPStatus

import pytest
import v3io.dataplane

from v3iofs import V3ioFS
from v3iofs.fs import _new_client

test_container = "bigdata"
test_dir = f"v3io-fs-test-{uuid.uuid4().hex}"

Obj = namedtuple("Obj", "path data")


@pytest.fixture
def fs():
    yield V3ioFS()


@pytest.fixture
def tmp_obj():
    user, ts = getuser(), datetime.now().isoformat()
    client = _new_client()

    path = f"{test_dir}/{user}-test-{ts}"
    body = f"test data for {user} at {ts}".encode()
    resp = client.put_object(test_container, path, body=body)
    assert resp.status_code == HTTPStatus.OK, "create failed"

    yield Obj(f"/{test_container}/{path}", body)

    client.delete_object(test_container, path)
    client.close()


@pytest.fixture
def new_file():
    _client, _path = None, ""

    def create_file(client, path, body=b""):
        nonlocal _client, _path

        _client, _path = client, path
        body = body or datetime.now().isoformat().encode("utf-8")
        out = client.put_object(test_container, path, body=body)
        out.raise_for_status()

    yield create_file

    _client.delete_object(
        container=test_container,
        path=_path,
        raise_for_status=v3io.dataplane.RaiseForStatus.never,
    )


@pytest.fixture
def client():
    client = _new_client()
    try:
        yield client
    finally:
        client.close()


tree_data = {
    "file1": b"file 1 data",
    "a": {
        "file1": b"file 1 data - a",
        "file2": b"file 2 data",
    },
    "b": {
        "file1": b"file 1 data - b",
    },
}
tree_root = f"/{test_container}/{test_dir}/tree"
Tree = namedtuple("Tree", "root data")


@pytest.fixture(scope="session")
def tree():
    fs = V3ioFS()
    create_tree(fs, tree_root, tree_data)

    yield Tree(tree_root, tree_data)

    fs.rm(tree_root, recursive=True)
    fs._client.close()


def create_tree(fs, prefix, tree):
    fs.mkdir(prefix)
    for name, value in tree.items():
        path = f"{prefix}/{name}"
        if isinstance(value, bytes):
            with fs.open(path, "wb") as out:
                out.write(value)
        elif isinstance(value, dict):
            fs.mkdir(path)
            create_tree(fs, path, value)
        else:
            raise TypeError(f"unknown value type: {type(value)!r}")



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import v3io
from fsspec.spec import AbstractBufferedFile
from v3io.dataplane import Client

from .path import split_container
from .utils import handle_v3io_errors


class V3ioFile(AbstractBufferedFile):
    def _fetch_range(self, start, end):
        client: Client = self.fs._client
        container, path = split_container(self.path)
        nbytes = end - start

        resp = client.get_object(
            container, path, offset=start, num_bytes=nbytes, raise_for_status=v3io.dataplane.RaiseForStatus.never
        )

        return handle_v3io_errors(resp, path)

    def _upload_chunk(self, final=False):
        """Write one part of a multi-block file upload

        Parameters
        ----------
        final: bool
            This is the last block, so should complete file, if
            self.autocommit is True.
        """
        body = self.buffer.getvalue()
        if not body:
            return

        client: Client = self.fs._client
        container, path = split_container(self.path)
        resp = client.put_object(
            container,
            path,
            body=body,
            append=True,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )

        if handle_v3io_errors(resp, path):
            # No need to clear self.buffer, fsspec does that
            return True

    def _initiate_upload(self):
        """Create remote file/upload"""
        self.fs.rm_file(self.path)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from v3iofs import path as v3path

split_cases = [
    # path, expected, raises
    ("", None, True),
    ("/", ("", ""), False),
    ("/a", ("a", ""), False),
    ("/a/b/c", ("a", "b/c"), False),
]


@pytest.mark.parametrize("path, expected, raises", split_cases)
def test_split_container(path, expected, raises):
    if raises:
        with pytest.raises(ValueError):
            v3path.split_container(path)
        return

    out = v3path.split_container(path)
    assert expected == tuple(out), "split"



from v3iofs.fs import _Cache


def test_put_and_get():
    cache = _Cache(10, 100)
    cache.put("k1", "v1")
    cache.put("k2", "v2")
    cache.put("k3", "v3.1")
    cache.put("k3", "v3.2")
    assert cache.get("k1") == "v1"
    assert cache.get("k2") == "v2"
    assert cache.get("k3") == "v3.2"


def test_invalidation_and_gc():
    cache = _Cache(10, 0)
    cache.put("k1", "v1")
    cache.put("k2", "v2")
    cache.put("k3", "v3.1")
    cache.put("k3", "v3.2")
    assert cache.get("k1") is None
    assert cache.get("k2") is None
    assert cache.get("k3") is None
    assert cache._cache == {}
    assert cache._expiry_to_key == []



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""v3iofs - An fsspec driver for v3io"""

__all__ = [
    "__version__",
    "V3ioFS",
    "V3ioFile",
]

__version__ = "0.1.16"

import fsspec

from .file import V3ioFile  # noqa: F401
from .fs import V3ioFS  # noqa: F401

if hasattr(fsspec, "register_implementation"):
    # TODO: Not sure about clobber=True
    fsspec.register_implementation("v3io", V3ioFS, clobber=True)
else:
    from fsspec.registry import known_implementations

    known_implementations["v3io"] = {
        "class": "v3iofs.V3ioFS",
        "err": "Please install v3iofs to use the v3io fileysstem class",
    }

    del known_implementations

del fsspec  # clear the module namespace



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from pathlib import Path

import dask.bag as db
import dask.dataframe as dd
import pytest
from conftest import test_container, test_dir

csv_data = b"""
name,item,price,quantity
Rick,Vodka,13.2,22
Jerry,Beer,34.2,300
Beth,Lettuce,1.2,13
Summer,M&M,2.2,7
Morty,Twix,1.7,5
"""

tests_root = Path(__file__).absolute().parent
test_pq = tests_root / "sanchez.pq"
with test_pq.open("rb") as fp:
    parquet_data = fp.read()

data_by_type = {
    "csv": csv_data,
    "pq": parquet_data,
}


def test_text(tmp_obj):
    uri = f"v3io://{tmp_obj.path}"
    file = db.read_text(uri)
    (data,) = file.compute(scheduler="single-threaded")
    data = data.encode("utf-8")
    assert tmp_obj.data == data, "bad data"


read_cases = [
    # read_fn, typ
    (dd.read_csv, "csv"),
    (dd.read_parquet, "pq"),
]


@pytest.mark.parametrize("read_fn, typ", read_cases)
def test_read(read_fn, typ, client, new_file):
    assert typ in data_by_type, f"unknown type: {typ!r}"
    data = data_by_type[typ]
    file_name = datetime.now().strftime("test-%Y%m%d%H%M.") + typ
    file_path = f"/{test_dir}/{file_name}"
    new_file(client, file_path, data)

    df = read_fn(f"v3io://{test_container}{file_path}")
    assert 4 == len(df.columns), "# of columns"
    assert 5 == len(df), "# of rows"



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime, timezone
from os.path import basename, dirname
from pathlib import Path

import fsspec
import pytest
from conftest import test_container, test_dir

from v3iofs import V3ioFS
from v3iofs.fs import parse_time
from v3iofs.path import split_container

path_types = [
    str,
    Path,
]


def makedir(path):
    fs = fsspec.filesystem("v3io")
    path += "/delete_me"
    with fs.open(path, "wb") as out:
        out.write("delete me".encode("UTF8"))
    fs.rm(path)


@pytest.mark.parametrize("path_cls", path_types)
def test_ls(fs: V3ioFS, new_file, path_cls):
    new_file(fs._client, f"{test_dir}/test-file")  # Make sure dir exists
    path = path_cls(f"/{test_container}/{test_dir}/")
    out = fs.ls(path)
    assert len(out) > 0, "nothing found"
    assert all(isinstance(p, dict) for p in out), "not dict"

    out = fs.ls(path, detail=False)
    assert len(out) > 0, "nothing found"
    assert all(isinstance(p, str) for p in out), "not string"


def test_ls_with_marker(fs: V3ioFS, new_file):
    for i in range(5):
        new_file(fs._client, f"{test_dir}/test_ls/test-file{i}")
    path = str(f"/{test_container}/{test_dir}/test_ls")

    out = fs.ls(path, detail=True, limit=3)
    assert len(out) == 5, "not all files returned"


def test_rm(fs: V3ioFS, tmp_obj):
    path = tmp_obj.path
    fs.rm(path)
    out = fs.ls(dirname(path), detail=False)
    assert path not in out, "not deleted"


def test_touch(fs: V3ioFS, tmp_obj):
    path = tmp_obj.path
    fs.touch(path)
    container, path = split_container(path)
    resp = fs._client.get_object(container, path)
    assert resp.body == b"", "not truncated"


now = datetime(2020, 1, 2, 3, 4, 5, 6789, tzinfo=timezone.utc)
parse_time_cases = [
    # value, expected, raises
    ("", None, True),
    (now.strftime("%Y-%m-%d"), None, True),
    (now.strftime("%Y-%m-%dT%H:%M:%S.%f%z"), now.timestamp(), False),
    (now.strftime("%Y-%m-%dT%H:%M:%S.%fZ"), now.timestamp(), False),
]


@pytest.mark.parametrize("value, expected, raises", parse_time_cases)
def test_parse_time(value, expected, raises):
    if raises:
        with pytest.raises(ValueError):
            parse_time(value)
        return

    out = parse_time(value)
    assert expected == out


def load_tree(fs, root):
    out = {}
    for entry in fs.ls(root, detail=True):
        name = basename(entry["name"])
        if entry["type"] == "file":
            with fs.open(entry["name"], "rb") as fp:
                out[name] = fp.read()
        elif entry["type"] == "directory":
            out[name] = load_tree(fs, f"{root}/{name}")
        else:
            raise TypeError(f'unknown entry type: {entry["type"]}')
    return out


def test_directory(fs, tree):
    out = load_tree(fs, tree.root)
    assert tree.data == out


def test_fsspec():
    fs = fsspec.filesystem("v3io")
    dirpath = f"/{test_container}/{test_dir}/fss"
    file_name = datetime.now().strftime("test_%f")
    filepath = f"{dirpath}/{file_name}.txt"
    with fs.open(filepath, "wb") as fp:
        fp.write(b"123")
    for prefix in ["", "v3io://"]:
        files = fs.ls(prefix + dirpath, detail=True)
        assert len(files) == 1, f"unexpected number of files {len(files)} in {prefix + dirpath}"
        assert fs.info(prefix + filepath)["type"] == "file", f"unexpected type in {prefix + dirpath}"
        assert fs.size(prefix + filepath) == 3, f"unexpected file size in {prefix + dirpath}"
    with fs.open(prefix + filepath) as fp:
        data = fp.read()
    assert data == b"123", "unexpected data"
    fs.rm(filepath)


def test_empty_directory():
    fs = fsspec.filesystem("v3io")
    path = f"/{test_container}/{test_dir}/test_empty_directory"
    makedir(path)
    try:
        result = fs.ls(path)
        assert result == []
    finally:
        fs.rmdir(path)


def test_directory_with_file():
    fs = fsspec.filesystem("v3io")
    dir_path = f"/{test_container}/{test_dir}/test_directory_with_file"
    file_path = f"/{test_container}/{test_dir}/test_directory_with_file/my_file"
    with fs.open(file_path, "wb") as out:
        out.write("test file".encode("UTF8"))
    try:
        result = fs.ls(dir_path)
        assert len(result) == 1
        assert result[0]["name"] == f"/{test_container}/{test_dir}/test_directory_with_file/my_file"
        assert result[0]["type"] == "file"
        assert result[0]["size"] == 9
    finally:
        fs.rm(file_path)
        fs.rmdir(dir_path)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import traceback
import weakref
from datetime import datetime, timezone
from os import environ
from threading import Lock
from urllib.parse import urlparse

import v3io
from fsspec.spec import AbstractFileSystem
from v3io.dataplane import Client

from .file import V3ioFile
from .path import path_equal, split_container, strip_schema, unslash
from .utils import handle_v3io_errors

_file_key = "key"
_dir_key = "prefix"


class _Cache:
    def __init__(self, capacity, cache_validity_seconds):
        self._cache = {}
        self._expiry_to_key = []
        self._capacity = capacity
        self._cache_validity_seconds = cache_validity_seconds

    def put(self, key, value):
        now = time.monotonic()
        expiry = now + self._cache_validity_seconds

        self._cache[key] = (expiry, value)
        self._expiry_to_key.append((expiry, key))

        # GC
        if len(self._cache) > self._capacity:
            self._gc(now)

    def get(self, key):
        lookup_result = self._cache.get(key)
        if lookup_result is None:
            return None
        expiry, value = lookup_result
        now = time.monotonic()
        if now <= expiry:
            return value

        self._gc(now)

        return None

    def delete_if_exists(self, key):
        self._cache.pop(key, None)

    def _gc(self, until):
        num_removed = 0
        for expiry, key in self._expiry_to_key:
            if expiry <= until or len(self._cache) > self._capacity:
                result = self._cache.get(key, None)
                if result:
                    cache_time, value = result
                    # cache entry may have been deleted and re-added
                    if cache_time == expiry:
                        del self._cache[key]

                num_removed += 1
            else:
                break
        self._expiry_to_key = self._expiry_to_key[num_removed:]


class V3ioFS(AbstractFileSystem):
    """File system driver to v3io

    Parameters
    ----------
    v3io_api: str
        API host name (or V3IO_API environment)
    v3io_access_key: str
        v3io access key (or V3IO_ACCESS_KEY from environment)
    cache_validity_seconds: int | str | None
        use caching for info(), with invalidation after cache_validity_seconds. Default is 2. Set to 0 to disable.
    cache_capacity: int | str | None
        limits the size of the cache. If cache_validity_seconds is not set, this parameter has no effect.
        Default is 128.
    **kw:
        Passed to fsspec.AbstractFileSystem
    """

    protocol = "v3io"

    def __init__(self, v3io_api=None, v3io_access_key=None, cache_validity_seconds=None, cache_capacity=None, **kw):
        # TODO: Support storage options for creds (in kw)
        super().__init__(**kw)
        self._client = _new_client(v3io_api, v3io_access_key)
        self._cache = None
        if cache_validity_seconds is None:
            cache_validity_seconds = 2
        if cache_capacity is None:
            cache_capacity = 128
        if cache_validity_seconds > 0:
            self._cache = _Cache(int(cache_capacity), int(cache_validity_seconds))
            self._cache_lock = Lock()
        weakref.finalize(self, lambda: self._client.close())

    def ls(self, path, detail=True, marker=None, **kwargs):
        """Lists files & directories under path"""

        path = strip_schema(path)
        container, path = split_container(path)
        if not container:
            return self._list_containers(detail)
        ext_out = []

        limit = kwargs.get("limit", None)

        while True:
            resp = self._client.get_container_contents(
                container=container,
                path=path,
                get_all_attributes=True,
                raise_for_status=v3io.dataplane.RaiseForStatus.never,
                limit=limit,
                marker=marker,
            )

            # Ignore 404's here
            if resp.status_code not in {200, 404}:
                raise Exception(f"{resp.status_code} received while accessing {path!r}")

            out = _resp_dirs(resp, container, detail) + _resp_files(resp, container, detail)

            if not marker:
                # first time in
                if not _has_data(resp):
                    return [self._ls_file(container, path, detail)]
                if not out:
                    return []

            ext_out.extend(out)
            if hasattr(resp.output, "next_marker") and resp.output.next_marker:
                marker = resp.output.next_marker
            else:
                break
        return ext_out

    def _ls_file(self, container, path, detail, marker=None):
        # '/a/b/c' -> ('/a/b', 'c')
        dirname, _, filename = path.rpartition("/")
        while True:
            resp = self._client.get_container_contents(
                container=container,
                path=dirname,
                get_all_attributes=True,
                raise_for_status=v3io.dataplane.RaiseForStatus.never,
                marker=marker,
            )

            full_path = f"/{container}/{path}"
            handle_v3io_errors(resp, full_path)
            contents = getattr(resp.output, "contents", [])
            objs = [obj for obj in contents if path_equal(obj.key, path)]
            if not objs:
                if hasattr(resp.output, "next_marker") and resp.output.next_marker:
                    marker = resp.output.next_marker
                else:
                    raise FileNotFoundError(full_path)
            else:
                break

        obj = objs[0]
        if not detail:
            return full_path
        return info_of(container, obj, _file_key)

    def _list_containers(self, detail):
        resp = self._client.get_containers(raise_for_status=v3io.dataplane.RaiseForStatus.never)
        handle_v3io_errors(resp, "containers")
        fn = container_info if detail else container_path
        return [fn(c) for c in resp.output.containers]

    def copy(self, path1, path2, **kwargs):
        ...  # FIXME

    def rmdir(self, path):
        if path and not path.endswith("/"):
            path += "/"
        self._rm(path)

    def _rm(self, path):
        container, path_without_container = split_container(path)
        if not container:
            raise ValueError(f"bad path: {path:r}")

        resp = self._client.delete_object(
            container=container,
            path=path_without_container,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )

        # Ignore 404's and 409's in delete
        if resp.status_code not in {200, 204, 404, 409}:
            raise Exception(f"{resp.status_code} received while accessing {path!r}")

        if self._cache:
            with self._cache_lock:
                self._cache.delete_if_exists(path)

    def touch(self, path, truncate=True, **kwargs):
        if not truncate:  # TODO
            raise ValueError("only truncate touch supported")

        path = strip_schema(path)
        container, path = split_container(path)
        resp = self._client.put_object(container, path, raise_for_status=v3io.dataplane.RaiseForStatus.never)

        handle_v3io_errors(resp, path)

    def info(self, path, **kw):
        """Details of entry at path

        Returns a single dictionary, with exactly the same information as
        ``ls`` would with ``detail=True``.

        Parameters
        ----------
        path: str
            Path to get info for
        **kw:
            Keyword arguments passed to `ls`

        Returns
        -------
        dict
            keys: name (full path in the FS), size (in bytes), type (file,
            directory, or something else) and other FS-specific keys.
        """
        path_with_container = strip_schema(path)

        if self._cache:
            with self._cache_lock:
                lookup_result = self._cache.get(path_with_container)
            if lookup_result:
                return lookup_result

        container, path_without_container = split_container(path_with_container)

        # First, we try to get the file's attributes, which will fail with a 404 if it's actually a directory.
        resp = self._client.get_item(
            container,
            path_without_container,
            attribute_names=["__size", "__mtime_secs", "__mtime_nsecs", "__mode", "__gid", "__uid"],
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )

        if resp.status_code == 200:
            mtime = int(resp.output.item["__mtime_secs"]) + int(resp.output.item["__mtime_nsecs"]) / 10**9
            entry = {
                "name": path_with_container,
                "type": "file",
                "size": resp.output.item["__size"],
                "mtime": mtime,
                "mode": resp.output.item["__mode"],
                "gid": resp.output.item["__gid"],
                "uid": resp.output.item["__uid"],
            }
            if self._cache:
                with self._cache_lock:
                    self._cache.put(path_with_container, entry)
            return entry
        elif resp.status_code == 404:
            pass  # The file may still be a directory.
        else:
            raise Exception(
                f"{resp.status_code} received while getting the attributes of {path_with_container!r}. "
                f"body={resp.body}, headers={resp.headers}"
            )

        # Check the existence of a directory at the provided path.
        resp = self._client.get_container_contents(
            container=container,
            path=path_without_container,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
            limit=0,
        )

        if resp.status_code == 200:
            entry = {"name": path_with_container, "size": 0, "type": "directory"}
            if self._cache:
                with self._cache_lock:
                    self._cache.put(path_with_container, entry)
            return entry
        elif resp.status_code == 404:
            raise FileNotFoundError(path_with_container)
        else:
            raise Exception(f"{resp.status_code} received while listing {path_with_container!r}")

    # Override to print the otherwise silenced exception.
    def isdir(self, path):
        """Is this entry directory-like?"""
        try:
            return self.info(path)["type"] == "directory"
        except IOError:
            traceback.print_exc()
            return False

    # Override to print the otherwise silenced exception.
    def isfile(self, path):
        """Is this entry file-like?"""
        try:
            return self.info(path)["type"] == "file"
        except FileNotFoundError:
            return False
        except BaseException:
            traceback.print_exc()
            return False

    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_type="readahead",
        cache_options=None,
        **kw,
    ):
        return V3ioFile(
            fs=self,
            path=path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_type=cache_type,
            cache_options=cache_options,
            **kw,
        )


def container_path(container):
    return f"/{container.name}"


def parse_time(creation_date):
    # '2020-03-26T09:42:57.504000+00:00'
    # '2020-03-26T09:42:57.71Z'
    i = creation_date.rfind("+")  # If not found will be -1, good for Z
    dt = datetime.strptime(creation_date[:i], "%Y-%m-%dT%H:%M:%S.%f")
    dt = dt.replace(tzinfo=timezone.utc)
    return dt.timestamp()


def container_info(container):
    return {
        "name": container.name,
        "size": None,
        "created": parse_time(container.creation_date),
    }


_missing = object()
_extra_obj_attrs = [
    # dest, src, convert
    ("created", "creating_time", parse_time),
    ("atime", "access_time", parse_time),
    ("mode", "mode", lambda v: int(v[1:], base=8)),  # '040755'
    ("gid", "gid", lambda v: int(v, 16)),
    ("uid", "uid", lambda v: int(v, 16)),
]


def obj_path(container, obj, name_key):
    path = unslash(getattr(obj, name_key))
    return f"/{container}/{path}"


def info_of(container_name, obj, name_key):
    info = {
        "name": obj_path(container_name, obj, name_key),
        "type": "file" if hasattr(obj, "size") else "directory",
        "size": getattr(obj, "size", 0),
        "mtime": parse_time(obj.last_modified),
    }

    for src, dest, conv in _extra_obj_attrs:
        val = getattr(obj, src, _missing)
        if val is _missing:
            continue
        info[dest] = conv(val)

    return info


def split_auth(url):
    """
    >>> split_auth('v3io://api_key:s3cr3t@domain.company.com')
    ('v3io://domain.company.com', 's3cr3t')
    >>> split_auth('v3io://domain.company.com')
    ('v3io://domain.company.com', '')
    """
    u = urlparse(url)
    if "@" not in u.netloc:
        return (url, "")

    auth, netloc = u.netloc.split("@", 1)
    if ":" not in auth:
        raise ValueError("missing : in auth")
    _, key = auth.split(":", 1)
    u = u._replace(netloc=netloc)
    return (u.geturl(), key)


def _resp_dirs(resp, container, detail):
    if not hasattr(resp.output, "common_prefixes"):
        return []

    objs = resp.output.common_prefixes
    if not detail:
        return [obj_path(container, obj, _dir_key) for obj in objs]

    return [info_of(container, obj, _dir_key) for obj in objs]


def _resp_files(resp, container, detail):
    if not hasattr(resp.output, "contents"):
        return []

    objs = resp.output.contents
    if not detail:
        return [obj_path(container, obj, _file_key) for obj in objs]

    return [info_of(container, obj, _file_key) for obj in objs]


def _has_data(resp):
    out = resp.output
    return hasattr(out, "common_prefixes") or hasattr(out, "contents")


def _new_client(v3io_api=None, v3io_access_key=None) -> Client:
    v3io_api = v3io_api or environ.get("V3IO_API")
    v3io_access_key = v3io_access_key or environ.get("V3IO_ACCESS_KEY")

    return Client(
        endpoint=v3io_api,
        access_key=v3io_access_key,
    )



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fsspec

from v3iofs import V3ioFS


def test_register():
    cls = fsspec.get_filesystem_class(V3ioFS.protocol)
    assert cls is V3ioFS, "not registered"

    fs = fsspec.filesystem("v3io")
    assert isinstance(fs, V3ioFS), f"bad object class - {fs.__class__}"



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime

import pandas as pd
from conftest import test_container, test_dir

csv_data = b"""
name,item,price,quantity
Rick,Vodka,13.2,22
Jerry,Beer,34.2,300
Beth,Lettuce,1.2,13
Summer,M&M,2.2,7
Morty,Twix,1.7,5
"""


def test_pandas(client, new_file):
    file_name = datetime.now().strftime("test-%Y%m%d%H%M.csv")
    file_path = f"/{test_dir}/{file_name}"
    new_file(client, file_path, csv_data)

    df = pd.read_csv(f"v3io://{test_container}{file_path}")
    assert 4 == len(df.columns), "# of columns"
    assert 5 == len(df), "# of rows"



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import find_packages, setup


def version():
    with open("v3iofs/__init__.py") as fp:
        for line in fp:
            if line.startswith("__version__"):
                _, version = line.split("=")
                return version.replace('"', "").strip()


def load_deps(file_name):
    """Load dependencies from requirements file"""
    deps = []
    with open(file_name) as fp:
        for line in fp:
            line = line.strip()
            if not line or line[0] == "#" or line.startswith("-r"):
                continue
            deps.append(line)
    return deps


install_requires = load_deps("requirements.txt")
tests_require = load_deps("dev-requirements.txt")

with open("README.md") as fp:
    long_desc = fp.read()

setup(
    name="v3iofs",
    version=version(),
    description="fsspec driver for v3io",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    author="Iguazio",
    author_email="yaronh@iguazio.com",
    license="MIT",
    url="https://github.com/v3io/v3io-fs",
    packages=find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
    ],
    tests_require=tests_require,
)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from urllib.parse import urlparse


def strip_schema(url):
    if not url:
        return url
    url = str(url)
    if "://" in url:
        url = urlparse(url).path
    return url.replace("\\", "/")  # fix in windows


def split_container(path):
    """Split path container & path

    >>> split_container('/bigdata/path/to/file')
    ['bigdata', 'path/to/file']
    """
    path = str(path)  # Might be pathlib.Path
    if not path:
        raise ValueError("empty path")
    if path == "/":
        return "", ""

    if path[0] == "/":
        path = path[1:]

    if "/" not in path:
        return path, ""  # container

    return path.split("/", maxsplit=1)


def unslash(s):
    """Remove optional slashes from the start/end."""
    return s.strip("/")


def _norm(p):
    p = re.sub("/+", "/", p)
    return "/" + p if p[0] != "/" else p


def path_equal(p1, p2):
    """Check that two paths are equal"""
    return _norm(p1) == _norm(p2)
